
#include <rpc/rpc.h>
#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <dirent.h>

#include <gssapi/gssapi.h>
#include <gssapi/gssapi_generic.h>
#include <gssapi/gssapi_krb5.h>


/* Python includes */
#include <Python.h>
#include "structmember.h"

/**********************************************************************/
/*  Buffer                                                            */
/**********************************************************************/

typedef struct {
	PyObject_HEAD
	size_t length;
	void  *data;
} BufferObj;

static PyObject *
Buffer_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
	BufferObj *self;

	self = (BufferObj *) type->tp_alloc(type, 0);
	if (self != NULL) {
		self->data = NULL;
		self->length = 0;
	}

	return (PyObject *)self;
}

static void
Buffer_dealloc(BufferObj *self)
{
	/* Note it is *NOT* our job to deallocate data */
	self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
Buffer_str(BufferObj *self)
{
	//printf("%u, %p\n",self->length, self->data);
	return PyString_FromFormat("{length: %i, pointer: %p}", 
				   (int)self->length, self->data);
}

#if 0
static PyObject *
Buffer_repr(BufferObj *self)
{
	printf("repr: as long: %x, as p: %p\n",(long) self->data, self->data); 
	return PyInt_FromLong( (long) self->data);
}
#endif

static PyObject *
Buffer_getlength(BufferObj *self, void *closure)
{
	return Py_BuildValue("i", (int)self->length);
}

static int
Buffer_setlength(BufferObj *self, PyObject *value, void *closure)
{
	/* XXX This should not be allowed to be changed */
	int length;
	char *message;

	if (value == NULL) {
		message = "Cannot delete the length attribute";
		goto error;
	}

	length = PyInt_AsLong(value);
	if (PyErr_Occurred()) {
		message = "The length attribute value must be an integer";
		goto error;
	}
	    
	if (length < 0) {
		message = "The length attribute must be positive";
		goto error;
	}
	
	self->length = (size_t)length;
	return 0;
error:
	PyErr_SetString(PyExc_TypeError, message);
	return -1;
}

static PyObject *
Buffer_getpointer(BufferObj *self, void *closure)
{
	return PyString_FromFormat("%p", self->data);
}

static int
Buffer_setpointer(BufferObj *self, PyObject *value, void *closure)
{
	PyErr_SetString(PyExc_TypeError, "Cannot set pointer attribute directly");
	return -1;
}

static PyGetSetDef Buffer_getseters[] = {
    {"pointer", 
     (getter)Buffer_getpointer, (setter)Buffer_setpointer,
     "Pointer to opaque data", NULL},
    {"length",
     (getter)Buffer_getlength, (setter)Buffer_setlength,
     "If needed, length of opaque data", NULL},
    {NULL}  /* Sentinel */
};

static PyObject *
Buffer_pulldata(BufferObj * self)
{
	if (self->length)
		return Py_BuildValue("s#", self->data, self->length);
	else 
		return PyInt_FromLong( (long) self->data);
}

static PyMethodDef Buffer_methods[] = {
    {"pulldata", (PyCFunction)Buffer_pulldata, METH_NOARGS,
     "Return data pointed to as a string"
    },
    {NULL}  /* Sentinel */
};


static PyTypeObject BufferType = {
	PyObject_HEAD_INIT(NULL)
	0,                             /* ob_size - NOT USED */
	"gssapi.Buffer",               /* tp_name */
	sizeof(BufferObj),             /* tp_basicsize */
	0,                             /* tp_itemsize */
	(destructor)Buffer_dealloc,    /* tp_dealloc */
	0,                             /* tp_print */
	0,                             /* tp_getattr */
	0,                             /* tp_setattr */
	0,                             /* tp_compare */
	(reprfunc)Buffer_str,          /* tp_repr */
	0,                             /* tp_as_number */
	0,                             /* tp_as_sequence */
	0,                             /* tp_as_mapping */
	0,                             /* tp_hash */
	0,                             /* tp_call */
	(reprfunc)Buffer_str,          /* tp_str */
	0,                             /* tp_getattro */
	0,                             /* tp_setattro */
	0,                             /* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,            /* tp_flags */
	"Pointer to opaque data",      /* tp_doc */
	0,		               /* tp_traverse */
	0,		               /* tp_clear */
	0,		               /* tp_richcompare */
	0,		               /* tp_weaklistoffset */
	0,		               /* tp_iter */
	0,		               /* tp_iternext */
	Buffer_methods,                /* tp_methods */
	0,                             /* tp_members */
	Buffer_getseters,              /* tp_getset */
	0,                             /* tp_base */
	0,                             /* tp_dict */
	0,                             /* tp_descr_get */
	0,                             /* tp_descr_set */
	0,                             /* tp_dictoffset */
	0,                             /* tp_init */
	0,                             /* tp_alloc */
	Buffer_new,                    /* tp_new */
};


/**********************************************************************/
/* Helper functions                                                   */
/**********************************************************************/

static BufferObj *
create_buffer(void)
{
	return (BufferObj *) BufferType.tp_new(&BufferType, NULL, NULL);
}

/**********************************************************************/
/* GSSAPI wrappers                                                    */
/**********************************************************************/

/* (major, minor, out, conf, qop) = unwrap(context, msg) */
static PyObject *
gssapi_unwrap(PyObject *self, PyObject *args)
{
	/* Input and output variables */
	BufferObj         *in_context=NULL;
	char              *in_message=NULL;
	int                msg_len=0;

	OM_uint32          major=0;
	OM_uint32          minor=0;
	OM_uint32          confidential=0;
	OM_uint32          qop=0;
	PyObject          *rv;

	/* Local variables */
	gss_buffer_desc	   out_msg, in_msg;

	if (!PyArg_ParseTuple(args, "O!s#",
			      &BufferType, &in_context,
			      &in_message, &msg_len))
		return NULL;

	in_msg.value = in_message;
	in_msg.length = msg_len;

	major = gss_unwrap(&minor, in_context->data, &in_msg, &out_msg,
			   &confidential, &qop);

	rv = Py_BuildValue("iis#ii",
			   major, minor, out_msg.value, out_msg.length,
			   confidential, qop);
	gss_release_buffer(&minor, &out_msg);
	return rv;
}

/* (major, minor, out, conf) = wrap(context, msg, conf=True, qop=0) */
static PyObject *
gssapi_wrap(PyObject *self, PyObject *args)
{
	/* Input and output variables */
	BufferObj         *in_context=NULL;
	char              *in_message=NULL;
	int                msg_len=0;
	OM_uint32          in_confidential=1;
	OM_uint32          qop=0;

	OM_uint32          major=0;
	OM_uint32          minor=0;
	OM_uint32          out_confidential=0;
	PyObject          *rv;

	/* Local variables */
	gss_buffer_desc	   out_msg, in_msg;

	if (!PyArg_ParseTuple(args, "O!s#|ii",
			      &BufferType, &in_context,
			      &in_message, &msg_len,
			      &in_confidential,
			      &qop))
		return NULL;

	in_msg.value = in_message;
	in_msg.length = msg_len;

	major = gss_wrap(&minor, in_context->data, in_confidential, qop,
			 &in_msg, &out_confidential, &out_msg);
	
	rv = Py_BuildValue("iis#i",
			   major, minor, out_msg.value, out_msg.length,
			   out_confidential);
	gss_release_buffer(&minor, &out_msg);
	return rv;
}


/* (major, minor, qop) = verifyMIC(context, message, token) */
static PyObject *
gssapi_verifyMIC(PyObject *self, PyObject *args)
{
	/* Input and output variables */
	BufferObj         *in_context=NULL;
	char              *in_message=NULL, *in_token=NULL;
	int                msg_len=0, tok_len=0;

	OM_uint32          qop=0;
	OM_uint32          major=0;
	OM_uint32          minor=0;

	/* Local variables */
	gss_buffer_desc	   token, message;

	if (!PyArg_ParseTuple(args, "O!s#s#",
			      &BufferType, &in_context,
			      &in_message, &msg_len,
			      &in_token, &tok_len))
		return NULL;

	message.value = in_message;
	message.length = msg_len;
	token.value = in_token;
	token.length = tok_len;

	major = gss_verify_mic(&minor, in_context->data, 
			       &message, &token, &qop);

	return Py_BuildValue("iii", major, minor, qop);
}

/* (major, minor, token) = getMIC(context, message, qop=0) */
static PyObject *
gssapi_getMIC(PyObject *self, PyObject *args)
{
	/* Input and output variables */
	BufferObj         *in_context=NULL;
	char              *in_message=NULL;
	int                msg_len=0;
	OM_uint32          qop=0;

	OM_uint32          major=0;
	OM_uint32          minor=0;
	PyObject          *rv;

	/* Local variables */
	gss_buffer_desc	   out_token, message;

	//printf("Entered getMIC\n");
	if (!PyArg_ParseTuple(args, "O!s#|i", 
			      &BufferType, &in_context,
			      &in_message, &msg_len,
			      &qop))
		return NULL;	

	message.value = in_message;
	message.length = msg_len;

	//printf("getMIC: using context %p\n", in_context->data);

	major = gss_get_mic(&minor, in_context->data, qop, &message,
			    &out_token);

	//printf("getMIC: major=%i, minor=%i\n", major, minor);
	//printf("getMIC: token length = %i\n",out_token.length);
	rv = Py_BuildValue("iis#",
			   major, minor, out_token.value, out_token.length);
	gss_release_buffer(&minor, &out_token);
	return rv;
}

/* (major, minor, token, context, mech, flags, time) =
   initSecContext(name, token=GSS_C_NO_BUFFER, context=GSS_C_NO_CONTEXT,
                  mech=krb5, cred=GSS_C_NO_CREDENTIAL, flags=GSS_C_MUTUAL_FLAG,
                  time=0, binding=NULL)
*/
static PyObject *
gssapi_initSecContext(PyObject *self, PyObject *args)
{
	/* STUB */

	static gss_OID_desc krb5oid = 
		{9, "\052\206\110\206\367\022\001\002\002"};
	/* Input and output variables */
	OM_uint32          major=0;
	OM_uint32          minor=0;
	BufferObj                            *in_name;
	BufferObj         *out_token=NULL;
	char                                 *in_token=NULL;
	int                                  in_tok_len=0;
	BufferObj         *out_context=NULL, *in_context=NULL;
	BufferObj         *out_mech=NULL,    *in_mech=NULL; /* gss_OID */ 
	BufferObj                            *in_cred=NULL;
        OM_uint32          out_flags=0,       in_flags=0;
        OM_uint32          out_time=0,        in_time=0;
	BufferObj                            *in_bind=NULL;

	/* Local variables */
	gss_cred_id_t      cred = GSS_C_NO_CREDENTIAL;
	gss_ctx_id_t	   ctx = GSS_C_NO_CONTEXT;
	gss_buffer_desc	   send_token, rcv_token, *rcv_tokenp=GSS_C_NO_BUFFER;

	/* XXX The problem with doing everything as BufferObj is that
	 * can not do any type checking.  Better to subtype for each.
	 */
	//printf("Entered initSecContext\n");
	if (!PyArg_ParseTuple(args, "O!|s#O!O!O!llO!", 
			      &BufferType, &in_name,
			      &in_token,   &in_tok_len,
			      &BufferType, &in_context,
			      &BufferType, &in_mech,
			      &BufferType, &in_cred,
			      &in_flags,
			      &in_time,
			      &BufferType, &in_bind))
		return NULL;

	/* Set defaults */
	if (in_token && in_tok_len) {
		rcv_tokenp = &rcv_token;
		rcv_token.value = in_token;
		rcv_token.length = in_tok_len;
	}
	if (in_cred) cred = in_cred->data;
	if (in_context) ctx = in_context->data;
	if (in_mech) {
		PyErr_SetString(PyExc_TypeError, "args past 'context' not implemented");
		return NULL;
	}
		
	/* Call function */
	major = gss_init_sec_context(&minor, cred, &ctx, in_name->data,
				     (gss_OID)&krb5oid, GSS_C_MUTUAL_FLAG,
				     0, NULL,
				     rcv_tokenp, NULL,
				     &send_token, &out_flags, NULL);

#if 0
	printf("major=%x\n",major);
	printf("GSS_S_CONTINUE_NEEDED = %x\n",GSS_S_CONTINUE_NEEDED);
	printf("GSS_S_FAILURE = %x\n",GSS_S_FAILURE);
	printf("Token length = %i\n",send_token.length);
	printf("Token value = %p\n",send_token.value);
#endif

	/* Translate return values */
	out_token = create_buffer();
	if (out_token==NULL)
		goto error;
	out_context = create_buffer();
	if (out_context==NULL)
		goto error;
	out_mech = create_buffer();
	if (out_mech==NULL)
		goto error;

	out_token->length = send_token.length;
	out_token->data = send_token.value;

	out_context->data = ctx;

	return Py_BuildValue("iiNNNii",
			     major, minor, out_token, out_context,
			     out_mech, out_flags, out_time);

 error:
	Py_XDECREF(out_context);
	Py_XDECREF(out_token);
	Py_XDECREF(out_mech);
	
	return NULL;
}

/* (major,minor,name) = importName(string, type=gss_nt_service_name) */
static PyObject *
gssapi_importName(PyObject *self, PyObject *args)
{
	char              *string, *type_name=NULL;
	int                len, type_len;
	OM_uint32          major=0, minor=0;
	gss_buffer_desc    sname;
	gss_OID_desc       type, *typep=(gss_OID)gss_nt_service_name;
	gss_name_t	   name = NULL;
	BufferObj         *name_py;
	PyObject          *rv;

	if (!PyArg_ParseTuple(args, "s#|s#",
			      &string, &len, 
			      &type_name, &type_len))
		return NULL;

	sname.value = string;
	sname.length = len;

	if (type_len && type_name) {
		type.length = type_len;
		type.elements = type_name;
		typep = &type;
	}

	major = gss_import_name(&minor, &sname, typep, &name);

	/* Turn name into a PyObect */
	name_py = create_buffer();
	if (name_py==NULL) return NULL;
	name_py->data = name;

	rv = Py_BuildValue("iiN", major, minor, (PyObject *)name_py);
	if (!rv) {
		Py_DECREF(name_py);
	}
	return rv;
}

/**********************************************************************/
/* Table of wrappers                                                  */
/**********************************************************************/

static PyMethodDef GssapiMethods[] = {
    {"importName", gssapi_importName, METH_VARARGS,
     "(major,minor,name) = importName(string, type=gss_nt_service_name)"},
    {"verifyMIC", gssapi_verifyMIC, METH_VARARGS,
     "(major, minor, qop) = verifyMIC(context, message, token)"},
    {"getMIC", gssapi_getMIC, METH_VARARGS,
     "(major, minor, token) = getMIC(context, message, qop=0)"},
    {"wrap", gssapi_wrap, METH_VARARGS,
     "(major, minor, out, conf) = wrap(context, msg, conf=True, qop=0)"},
    {"unwrap", gssapi_unwrap, METH_VARARGS,
     "(major, minor, out, conf, qop) = unwrap(context, msg)"},
    {"initSecContext", gssapi_initSecContext, METH_VARARGS,
     "(major, minor, token, context, mech=0, flags, time=0) = \n"
     "initSecContext(name, token=NULL, context=NULL, mech=krb5, cred=NULL,\n"
     "               flags=GSS_C_MUTUAL_FLAG, time=0, binding=NULL)"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};



/**********************************************************************/
/* Module initialization                                              */
/**********************************************************************/


/* Module initialization...the only nonstatic entity */
#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
initgssapi(void)
{

	PyObject *m;

	/* Initialize new variable types */
	if (PyType_Ready(&BufferType) < 0)
		return;

	/* This is the critical line */
	m = Py_InitModule3("gssapi", GssapiMethods,"Wrapper for C gssapi routines");
	if (m==NULL) return;

	/* Make sure type stays in existance as long as module does */
	Py_INCREF(&BufferType);
	PyModule_AddObject(m, "Buffer", (PyObject *)&BufferType);

	PyModule_AddIntConstant(m, "GSS_S_COMPLETE", 0);
	PyModule_AddIntConstant(m, "GSS_S_CONTINUE_NEEDED", 1);
	PyModule_AddIntConstant(m, "GSS_S_FAILURE", 0xd0000);
}


